@@ -1,6 +1,6 @@
 //! Parses a stream of tokens into an AST.
 
-use std::borrow::Cow;
+use std::{borrow::Cow, fmt::Debug, marker::PhantomData};
 
 use tagged_box::TaggableContainer;
 //use smallvec::SmallVec;
@@ -8,14 +8,11 @@ use thiserror::Error;
 use rustc_hash::FxHashMap;
 use smartstring::alias::CompactString;
 
-use crate::{
-    lexer::{Literal, PeekableTokenStream, Symbol, Token, TokenContent, TokenStream},
-    values::*
-};
+use crate::{lexer::{Literal, PeekableTokenStream, Symbol, Token, TokenContent, TokenStream}, values::*};
 
 /// Parses a `TokenStream` into an AST.
 pub fn parse_expression(tokens: &mut PeekableTokenStream<impl TokenStream>) -> Result<Expression, ParseError> {
-    let expr = match tokens.next() {
+    let mut expr = match tokens.next() {
         // Empty case? Error!
         None => return Err(ParseError::Empty),
         
@@ -45,6 +42,34 @@ pub fn parse_expression(tokens: &mut PeekableTokenStream<impl TokenStream>) -> R
             content: TokenContent::Symbol(Symbol::CurlyLeft), ..
         }) => parse_map(tokens).map(|v| Expression::Value(ValContainer::from(v)))?,
         
+        // Global Variable!
+        Some(Token {
+            content: TokenContent::Symbol(Symbol::At), ..
+        }) => match tokens.next() {
+            Some(Token {
+                content: TokenContent::Literal(Literal::Str(s)), ..
+            }) => Expression::Value(ValContainer::from(GlobalVar(s))),
+            Some(t) => return Err(ParseError::ExpectButGot("a global variable name".into(), format!("{}", t).into())),
+            None => return Err(ParseError::ExpectButEnd("a global variable name")),
+        },
+        
+        // Local Variable!
+        Some(Token {
+            content: TokenContent::Symbol(Symbol::DollarSign), ..
+        }) => match tokens.next() {
+            Some(Token {
+                content: TokenContent::Symbol(Symbol::DollarSign), ..
+            }) => Expression::Value(ValItem::ResultVar(PhantomData::default()).into()),
+            Some(Token {
+                content: TokenContent::Literal(Literal::Str(s)), ..
+            }) => Expression::Value(ValContainer::from(LocalVar(s))),
+            Some(Token {
+                content: TokenContent::Literal(Literal::Int(i)), ..
+            }) => Expression::Value(ValContainer::from(LocalVar(i.to_string().into()))),
+            Some(t) => return Err(ParseError::ExpectButGot("a local variable name".into(), format!("{}", t).into())),
+            None => return Err(ParseError::ExpectButEnd("a local variable name")),
+        },
+        
         // Doubledot? Invalid!
         Some(Token {
             content: TokenContent::Symbol(Symbol::DoubleDot), ..
@@ -58,13 +83,65 @@ pub fn parse_expression(tokens: &mut PeekableTokenStream<impl TokenStream>) -> R
         }
     };
     
+    if let Some(token) = tokens.peek() {
+        match token {
+            // Dot? Subset access!
+            Token {
+                content: TokenContent::Symbol(Symbol::Dot), ..
+            } => {
+                drop(tokens.next()); // drop the dot
+                let mut get = Invoke {
+                    name: "idx".into(),
+                    pos_args: vec![expr],
+                    nom_args: Default::default(),
+                };
+                
+                if let Some(Token {
+                    content: TokenContent::Symbol(Symbol::QuestionMark), ..
+                }) = tokens.peek() {
+                    get.name = "idxn".into();
+                    drop(tokens.next()); // drop the question-mark
+                }
+                
+                let inner = parse_expression(tokens)?;
+                get.pos_args.push(inner);
+                
+                expr = Expression::Invoke(get.into());
+            },
+            
+            // QuestionMark? Existence check!
+            Token {
+                content: TokenContent::Symbol(Symbol::QuestionMark), ..
+            } => {
+                drop(tokens.next()); // drop the dot
+                expr = Expression::Invoke(Invoke {
+                    name: "exists".into(),
+                    pos_args: vec![expr],
+                    nom_args: Default::default(),
+                }.into());
+            },
+            
+            // Tilde? Relation!
+            Token {
+                content: TokenContent::Symbol(Symbol::Tilde), ..
+            } => {
+                drop(tokens.next()); // drop the dot
+                let to = parse_expression(tokens)?;
+                expr = Expression::Invoke(Invoke {
+                    name: "rel".into(),
+                    pos_args: vec![expr, to],
+                    nom_args: Default::default(),
+                }.into());
+            },
+            
+            // Ignore everything else...
+            _ => ()
+        }
+    }
+    
     // postfix stuff!
     // TODO: Ranges
     // TODO: Units
-    // TODO: RelativeTo
-    // TODO: Exists
-    // TODO: Dot-Indexing
-    // TODO: Box-Indexing
     // etc. etc.
     
     Ok(expr)
@@ -108,6 +185,15 @@ pub fn parse_command(
     
     loop {
         match tokens.peek().cloned() {
+            Some(Token {
+                content: TokenContent::Symbol(s), ..
+            })
+                if terminator.map_or(false, |t| t == s)
+            => {
+                // We do NOT drop the terminator here...
+                break // natural end of command, due to terminator.
+            },
+            
             Some(Token {
                 content: TokenContent::Symbol(Symbol::DoubleDot), ..
             }) => {
@@ -117,23 +203,71 @@ pub fn parse_command(
                 break; // natural end of command, due to subcommand
             },
             
+            Some(Token {
+                content: TokenContent::Symbol(Symbol::Ampersand), ..
+            }) => {
+                drop(tokens.next());
+                
+                let previous = std::mem::replace(&mut cmd, Invoke {
+                    name: "if-then".into(),
+                    pos_args: Default::default(),
+                    nom_args: Default::default(),
+                });
+                
+                cmd.pos_args.push(previous.into());
+                
+                let subcommand = parse_command(tokens, None)?;
+                cmd.pos_args.push(subcommand.into());
+                break; // natural end of command, due to subcommand
+            },
+            
             Some(Token {
                 content: TokenContent::Symbol(s), ..
             })
                 if s.is_end_delimiter()
-                || terminator.map_or(false, |t| t == s)
             => {
                 drop(tokens.next());
-                break // natural end of command, due to delimiter or terminator.
+                break // natural end of command, due to delimiter.
             },
             
-            None => if terminator.is_some() {
-                // abnormal end of command, due to missing terminator.
-                return Err(ParseError::ExpectButEnd("a delimiter"))
-            } else {
-                break; // natural end of command, due to EOS.
+            Some(Token {
+                content: TokenContent::Symbol(Symbol::Pipe), ..
+            }) => {
+                let previous = std::mem::replace(&mut cmd, Invoke {
+                    name: "pipe".into(),
+                    pos_args: Default::default(),
+                    nom_args: Default::default(),
+                });
+                
+                cmd.pos_args.push(previous.into());
+                
+                // parse further commands as long as there are pipe symbols
+                while let Some(Token { content: TokenContent::Symbol(Symbol::Pipe), .. }) = tokens.peek() {
+                    drop(tokens.next()); // drop the pipe symbol
+                    
+                    if tokens.peek().is_none() {
+                        return Err(ParseError::ExpectButEnd("another command in the pipe"));
+                    }
+                    
+                    if let Some(Token { content: TokenContent::Symbol(Symbol::QuestionMark), .. }) = tokens.peek() {
+                        drop(tokens.next()); // drop the question-mark
+                    } else {
+                        cmd.pos_args.push(Invoke {
+                            name: "nonull".into(),
+                            pos_args: vec![Expression::Value(ValContainer::from(PhantomData::<Result<(),()>>::default()))],
+                            ..Default::default()
+                        }.into());
+                    }
+                    
+                    let subcommand = parse_command(tokens, Some(Symbol::Pipe))?;
+                    cmd.pos_args.push(subcommand.into());
+                }
+                
+                break; // natural end of command, no more pipes.
             },
             
+            None => break, // natural end of command, due to EOS.
+            
             // Attempt parsing arguments...
             Some(_) => {
                 // ...starting with what may just be a expression...
@@ -147,7 +281,7 @@ pub fn parse_command(
                         let lexpr = match expr {
                             Expression::Value(val) => match val.into_inner() {
                                 ValItem::String(s) => s,
-                                _ => return Err(ParseError::ExpectButGot("a parameter name".into(), "something else".into())),
+                                val => return Err(ParseError::ExpectButGot("a parameter name".into(), format!("{}", val).into())),
                             },
                             _ => return Err(ParseError::ExpectButGot("a parameter name".into(), "a symbol or command".into())),
                         };
@@ -277,14 +411,14 @@ impl From<Invoke> for Expression {
 impl std::fmt::Display for Expression {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         match self {
-            Expression::Value(l) => l.fmt(f),
+            Expression::Value(l) => std::fmt::Display::fmt(l, f),
             Expression::Invoke(c) => write!(f, "({})", c),
         }
     }
 }
 
 /// A command to be evaluated.
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone, Default, PartialEq)]
 pub struct Invoke {
     /// The name of the command.
     pub name: CompactString,
@@ -317,7 +451,7 @@ pub enum ParseError {
     Empty,
     
     /// There was a character that could not be tokenized/lexed.
-    #[error("Unrecognized character at {0}: {1}")]
+    #[error("Unrecognized token at {0}: {1}")]
     Unrecognized(usize, String),
     
     /// The stream of tokens ended unexpectedly.
@@ -354,17 +488,31 @@ mod tests {
     
     #[test]
     fn should_succeed() -> Result<(), ParseError> {
-        eprintln!("-> {}", parse_command(&mut tokenize("= 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("+ 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("- 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("* 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("/ 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("test 1 2 3"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("test 1 2 3 a=4"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("mul 2 (+ 1 2 3)"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("test foo: bar baz"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("test [1 2 3 4 5]"), None)?);
-        eprintln!("-> {}", parse_command(&mut tokenize("test {a = 1, b=2, c=-3}"), None)?);
+        fn chk(input: &str) -> Result<(), ParseError> {
+            let output = parse_command(&mut tokenize(input), None)?;
+            eprintln!("INPUT:  {},\t PARSED:  {}", input, output);
+            Ok(())
+        }
+        
+        chk("= 1 2 3")?;
+        chk("+ 1 2 3")?;
+        chk("- 1 2 3")?;
+        chk("* 1 2 3")?;
+        chk("/ 1 2 3")?;
+        chk("test 1 2 3")?;
+        chk("test 1 2 3 a=4")?;
+        chk("mul 2 (+ 1 2 3)")?;
+        chk("test foo: bar baz")?;
+        chk("test [1 2 3 4 5]")?;
+        chk("test {a = 1, b=2, c=-3}")?;
+        chk("testA 1 2 3 | testB 4 5 6 | testC 7 8 9")?;
+        chk("maybe-null |? accepts-null")?;
+        chk("conditional & execution")?;
+        chk("echo \"Hello, World!\" @s.chat ")?;
+        chk("tp @a 0 0 0")?;
+        chk("tp @a @world.spawn")?;
+        chk("tp @a 0 100 0 rel=@self")?;
+        chk("for @a: tp [0 100 0]~$$")?;
         Ok(())
     }
     
